package GGI;

=head1 NAME

GGI - Perl extension for manipulating LibGGI visuals/inputs

=head1 SYNOPSIS

  use GGI;

  $visual = new GGI;

  $mode = $visual->checkmode()
  $mode = $visual->checkmode($mode)
  $mode = $visual->checkmode($sx, $sy, $frames, $graphtype);
  $mode = $visual->checkmode($visx, $visy, $virtx, $virty, $graphtype);
  $mode = $visual->checkmode($visx, $visy, $virtx, $virty, 
			     $cellx, $celly, $graphtype);

  $mode = $visual->setmode()
  $mode = $visual->setmode($mode)
  $visual->setmode($sx, $sy, $frames, $graphtype);
  $visual->setmode($visx, $visy, $virtx, $virty, $graphtype);
  $visual->setmode($visx, $visy, $virtx, $virty, 
		   $cellx, $celly, $graphtype);

  $mode = $visual->mode;
  $visual->getmode($mode);

  $visual->fg($pixelval);
  $visual->bg($pixelval);
  $visual->clip($left, $top, $right, $bottom);
  ($left, $top, $right, $bottom) = $visual->clip;

  $visual->palette();
  $colors = $visual->palette(undef);
  $visual->palette($colors);
  $visual->palette($start, $colors);
  $visual->palette($start, $numcolors, $colors);
  $visual->palette(\@colors);
  $visual->palette($channels, \@colors);
  $visual->palette($channels, $start, \@colors);
  $visual->palette($channels, $start, $numcolors, \@colors);

  $pixels = $visual->convert($colors);
  $pixels = $visual->convert($start_color, $colors);
  $pixels = $visual->convert($start_color, $num, $colors);
  $colors = $visual->convert($pixels);
  $colors = $visual->convert($pixel);

  $visual->fillscreen();

  $visual->pixel($destx, $desty);
  $visual->pixel($destx, $desty, $pixel);
  $visual->pixel($pixel, $srcx, $srcy);
  $visual->pixel($destx, $desty, $srcvis, $srcx, $srcy);

  $visual->hline($destx, $desty, $width);
  $visual->hline($destx, $desty, $width, $pixels);
  $visual->hline($pixels, $width, $srcx, $srcy);
  $visual->hline($destx, $desty, $width, $srcvis, $srcx, $srcy);

  $visual->vline($destx, $desty, $height);
  $visual->vline($destx, $desty, $height, $pixels);
  $visual->vline($pixels, $height, $srcx, $srcy);
  $visual->vline($destx, $desty, $height, $srcvis, $srcx, $srcy);

  $visual->line($x1, $y1, $x2, $y2);

  $visual->box($destx, $desty, $w, $h);
  $visual->box($destx, $desty, $w, $h, $pixels);
  $visual->box($pixels, $w, $h, $srcx, $srcy);
  $visual->box($destx, $desty, $w, $h, $srcx, $srcy);
  $visual->box($destx, $desty, $w, $h, $srcvis, $srcx, $srcy);

  $visual->puts($destx, $desty, $string);

  $visual->flush();
  $visual->flush($x, $y);

  $ev = $visual->read;
  $ev = $visual->read($mask);
  $visual->read(undef, $ev);
  $visual->read($mask, $ev);

  $mask = $visual->poll($timeout);
  $mask = $visual->poll($timeout, $mask);

  $howmany_queued = $visual->queued;

  $mask = $visual->mask;
  $visual->mask($mask);

=head1 DESCRIPTION

This XS extension encapsulates the LibGGI graphics API in a perl 
class.  Using this extension implies loading and initializing LibGGI;
this is done before the calling script is compiled.  LibGGI will
be properly shut down upon the termination of the calling script.
Any LibGGI structures allocated will be freed when their coinciding
perl objects are destroyed.  As a result, Perl graphics programs can 
be very short, simple and convenient.

=cut

# protos of stuff I haven't implemented yet.
#    int ggiSetOrigin(ggi_visual_t vis,int x,int y);
#    int ggiGetOrigin(ggi_visual_t vis,int *x,int *y);
#    int ggiSetDisplayFrame(ggi_visual_t vis, int frameno);
#    int ggiSetReadFrame(ggi_visual_t vis, int frameno);
#    int ggiSetWriteFrame(ggi_visual_t vis, int frameno);
#    int ggiGetDisplayFrame(ggi_visual_t vis);
#    int ggiGetReadFrame(ggi_visual_t vis);
#    int ggiGetWriteFrame(ggi_visual_t vis);

#    void ggiPanic(const char *format,...);
#    int        ggiSetFlags(ggi_visual_t vis,ggi_flags flags);
#    ggi_flags  ggiGetFlags(ggi_visual_t vis);
#    const ggi_pixelformat *ggiGetPixelFormat(ggi_visual_t vis);

#    int ggiGetGamma(ggi_visual_t vis,ggi_float *r,ggi_float *g,ggi_float *b);
#    int ggiSetGamma(ggi_visual_t vis,ggi_float r,ggi_float g,ggi_float b);
#    int ggiGetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);
#    int ggiSetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);

#    int ggiGetCharSize(ggi_visual_t vis,int *width,int *height);
#    int ggiKbhit(ggi_visual_t vis);
#    int ggiGetc(ggi_visual_t vis);



use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use GGI::Mode;
use GGI::Coord;
use GGI::GraphType;
use GGI::Color;
use GII;
use overload '""' => \&stringify;

require Exporter;
require DynaLoader;
require AutoLoader;

@ISA = qw(Exporter AutoLoader DynaLoader GII);
@EXPORT = qw();
$VERSION = '0.03';

bootstrap GGI $VERSION;

die "Cannot Initialize LibGGI!  Bailing..." if ggiInit();

# Preloaded methods

=head1 METHODS

GGI objects support the following methods.  Methods that take 
a coordinate or a width/height dimension will accept either two 
perl scalars or one GGI::Coord object in the appropriate position.

=head2 new

The object constructor calls LibGGI's ggiOpen function to create a
new LibGGI "visual" object which is represented by a perl GGI::
blessed reference object.  Arguments to the constructor will be
passed to ggiOpen (IMPLEMENTATION NOTE: currently only the first 
argument, a LibGGI display string, is sent).  The default is to 
pass no arguments, which will open the default display for the 
given environment.  If the visual cannot be opened, a scalar
undef is returned instead of a GGI:: object.

=cut

sub new {
    my ($class) = shift();
    my ($this) = {};
    if (@_ + 0) {
	$this->{display} = [@_];
	$this->{obj} = ggiOpen(@_);
    }
    else {
	$this->{display} = [$ENV{GGI_DISPLAY}, "(default)"];
	$this->{obj} = ggiOpen(undef);
    }
    return undef unless $this->{obj};
    bless $this, $class;
    return $this;
}

sub DESTROY {
    my ($this) = shift();
    ggiClose($this->{obj});
}

=head2 checkmode

The checkmode method will attempt to check whether the GGI:: object's 
visual would succeed a setmode operation based on the arguments passed.  

The behavior of this method varies depending on the number and type of the 
arguments that it is given.

If it is given no parameters or one undefined parameter, LibGGI's 
ggiCheckMode function will be called with values that will cause 
the default mode to be selected.  A GGI::Mode object is returned 
representing a suggested mode; if the mode check failed, the 
returned GGI::Mode object's "status" method contains the LibGGI error 
code returned by ggiCheckMode, GGI_OK (zero) for success. Note that a 
GGI::Mode object will return the value of its "status" method when evaluated 
in a numerical or boolean context.  The GGI::Mode's "lastused" method
is set to point at the invoking GGI:: object.

If it is given one parameter which is a GGI::Mode object, LibGGI's
ggiCheckMode function will be called using a copy of the GGI::Mode object 
as a "requested mode".  The contents of the GGI::Mode passed in the
parameter list will not be altered.  A GGI::Mode object is returned 
representing a suggested mode; The returned GGI::Mode object's "status" 
method contains the LibGGI error code returned by ggiCheckMode, GGI_OK (zero) 
for success.  Note that a GGI::Mode object will return the value of its 
"status" method when evaluated in a numerical or boolean context.
The GGI::Mode's "lastused" method is set to point at the invoking GGI:: object.

If it is given a size (as either two scalars or a GGI::Coord object), 
a number of frames, and a GGI::GraphType (either as a GGI::Graphtype 
or as a regular perl scalar which will be evaluated in a numeric context), 
LibGGI's ggiCheckSimpleMode function will be called.  A GGI::Mode object 
is returned representing a suggested mode; The returned GGI::Mode object's 
"status" method contains the LibGGI error code returned by ggiCheckMode, 
GGI_OK (zero) for success.  Note that a GGI::Mode object will return the 
value of its "status" method when evaluated in a numerical or boolean context.
The GGI::Mode's "lastused" method is set to point at the invoking 
GGI:: object.

If it is given visible and virtual sizes (with either represented 
as either two scalars or a GGI::Coord) and a GGI::GraphType (either as a 
GGI::Graphtype or as a regular perl scalar which will be evaluated in a 
numeric context), LibGGI's ggiCheckGraphMode function will be called.  A 
GGI::Mode object is returned representing a suggested mode; The returned 
GGI::Mode object's "status" method contains the LibGGI error code returned 
by ggiCheckMode, GGI_OK (zero) for success.  Note that a GGI::Mode object will 
return the value of its "status" method when evaluated in a numerical 
or boolean context.  The GGI::Mode's "lastused" method is set to point at 
the invoking GGI:: object.

Finally, if it is given visible, virtual, and character font cell sizes
(with any of them represented as either two scalars or a GGI::Coord)
and a GGI::Graphtype (either as a GGI::Graphtype or as a regular perl 
scalar which will be evaluated in a numeric context), LibGGI's 
ggiCheckTextMode function will be called.  A GGI::Mode object is returned 
representing a suggested mode; The returned GGI::Mode object's "status" 
method contains the LibGGI error code returned by ggiCheckMode, GGI_OK (zero) 
for success.  Note that a GGI::Mode object will return the value of its 
"status" method when evaluated in a numerical or boolean context.
The GGI::Mode's "lastused" method is set to point at the invoking GGI:: object.

The value undef() may be used instead of the LibGGI macro values 
GGI_AUTO and GT_AUTO, or the macro values may be imported by explicitly 
calling use() on the GGI::Mode and GGI::GraphType modules, respectively
and used as they are in LibGGI.

=cut

sub checkmode {
    my($items) = @_ - 1;
    my($this) = shift();
    my($arg) = shift();
    my($res) = new GGI::Mode;
    $res->lastused($this);
    if ($items <= 1) {
	if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Mode')) {
	    $res->frames($arg->frames);
	    $res->dpp($arg->dpp);
	    $res->size($arg->size);
	    $res->virt($arg->virt);
	    $res->visible($arg->visible);
	    $res->graphtype($arg->graphtype);
	}
	else {
	    $res->frames(GGI_AUTO);
	    $res->dpp(GGI_AUTO, GGI_AUTO);
	    $res->size(GGI_AUTO, GGI_AUTO);
	    $res->virt(GGI_AUTO, GGI_AUTO);
	    $res->visible(GGI_AUTO, GGI_AUTO);
	    $res->graphtype(GT_AUTO);
	}
	$res->status($this->_check_mode($this->{obj}, $res->{obj}));
	return $res;
    }
    # At least two parameters.  Get visible size.
    my($visx, $visy);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($visx, $visy) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($visx, $visy) = ($arg, shift());
    }
    $visx = GGI_AUTO unless defined($visx);
    $visy = GGI_AUTO unless defined($visy);
    if ($items == 4) {
	my($frames) = shift();
	$frames = GGI_AUTO unless defined($frames);
	my($graphtype) = shift();
	$graphtype = GT_AUTO unless defined($graphtype);
	$graphtype = $graphtype->graphtype()
	    if ref($graphtype) 
		and UNIVERSAL::isa($graphtype, 'GGI::GraphType');
	$res->status($this->_check_simple_mode($this->{obj}, 
					       $visx, $visy, $frames, 
					       $graphtype, $res->{obj}));
	return $res;
    }
    # More than four parameters, get virtual size
    my($virtx, $virty);
    $arg = shift();
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($virtx, $virty) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($virtx, $virty) = ($arg, shift());
    }
    $virtx = GGI_AUTO unless defined($virtx);
    $virty = GGI_AUTO unless defined($virty);
    $arg = shift();
    if ($items == 5) {
	$arg = GT_AUTO unless defined($arg);
	$arg = $arg->graphtype() if ref($arg) 
	    and UNIVERSAL::isa($arg, 'GGI::GraphType');
	$res->status($this->_check_graph_mode($this->{obj}, $visx, $visy, 
					      $virtx, $virty, $arg, 
					      $res->{obj}));
	return($res);
    }
    # More than five parameters, get character cell size
    my($cellx, $celly);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($cellx, $celly) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($cellx, $celly) = ($arg, shift());
    }
    $cellx = GGI_AUTO unless defined($cellx);
    $celly = GGI_AUTO unless defined($celly);
    $arg = shift();
    $arg = GT_AUTO unless defined($arg);
    $arg = $arg->graphtype 
	if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::GraphType'));
    $res->status($this->_check_text_mode($this->{obj}, 
					 $visx, $visy, $virtx, $virty, 
					 $cellx, $celly, 
					 $arg, $res->{obj}));
    return $res;
}


=head2 setmode

The setmode method will attempt to set the GGI:: object's visual's 
mode based on the arguments passed.  

The behavior of this method varies depending on the number and type of the 
arguments that it is given.

If it is given no parameters or one undefined parameter, LibGGI's 
ggiSetMode function will be called with values that will cause 
the default mode to be selected.  A GGI::Mode object is returned 
representing a suggested mode (or the actual mode set if the attempt 
suceeded); if the mode set failed, the returned GGI::Mode object's 
"status" method contains the LibGGI error code returned by ggiSetMode.  
Note that a GGI::Mode object will return the value of its "status" 
method when evaluated in a numerical or boolean context.

If it is given one parameter which is a GGI::Mode object, LibGGI's
ggiSetMode function will be called using a copy of the GGI::Mode object 
as a "requested mode".  The contents of the GGI::Mode passed in the
parameter list will not be altered.  A GGI::Mode object is returned 
representing a suggested mode (or the actual mode set if the attempt 
suceeded); The returned GGI::Mode object's "status" method contains
the LibGGI error code returned by ggiSetMode, GGI_OK (zero) for success.  Note 
that a GGI::Mode object will return the value of its "status" method 
when evaluated in a numerical or boolean context.

If it is given a size (as either two scalars or a GGI::Coord object), 
a number of frames, and a GGI::GraphType (either as a GGI::Graphtype 
or as a regular perl scalar which will be evaluated in a numeric context), 
LibGGI's ggiSetSimpleMode function will be called and the LibGGI error
code, GGI_OK (zero) for success, will be returned as a perl scalar.

If it is given visible and virtual sizes (with either represented 
as either two scalars or a GGI::Coord) and a GGI::GraphType (either as a 
GGI::Graphtype or as a regular perl scalar which will be evaluated in a 
numeric context), LibGGI's ggiSetGraphMode function will be called and 
the LibGGI error code, GGI_OK (zero) for success, will be returned as a 
perl scalar.

Finally, if it is given visible, virtual, and character font cell sizes
(with any of them represented as either two scalars or a GGI::Coord)
and a GGI::Graphtype (either as a GGI::Graphtype or as a regular perl 
scalar which will be evaluated in a numeric context), LibGGI's 
ggiSetTextMode function will be called and the LibGGI error code, GGI_OK 
(zero) for success, will be returned as a perl scalar.  

The value undef() may be used instead of the LibGGI macro values 
GGI_AUTO and GT_AUTO, or the macro values may be imported by explicitly 
calling use() on the GGI::Mode and GGI::GraphType modules, respectively
and used as they are in LibGGI.

=cut

sub setmode {
    my($items) = @_ - 1;
    my($this) = shift();
    my($arg) = shift();
    my($res) = new GGI::Mode;
    if ($items <= 1) {
	if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Mode')) {
	    $res->frames($arg->frames);
	    $res->dpp($arg->dpp);
	    $res->size($arg->size);
	    $res->virt($arg->virt);
	    $res->visible($arg->visible);
	    $res->graphtype($arg->graphtype);
	}
	else {
	    $res->frames(GGI_AUTO);
	    $res->dpp(GGI_AUTO, GGI_AUTO);
	    $res->size(GGI_AUTO, GGI_AUTO);
	    $res->virt(GGI_AUTO, GGI_AUTO);
	    $res->visible(GGI_AUTO, GGI_AUTO);
	    $res->graphtype(GT_AUTO);
	}
	$res->lastused($this);
	$res->status($this->_set_mode($this->{obj}, $res->{obj}));
	delete($this->{pfmt}) unless $res; # Force new ggi_pixel_format alloc
	return $res;
    }
    # At least two parameters.  Get visible size.
    my($visx, $visy);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($visx, $visy) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($visx, $visy) = ($arg, shift());
    }
    $visx = GGI_AUTO unless defined($visx);
    $visy = GGI_AUTO unless defined($visy);
    if ($items == 4) {
	my($frames) = shift();
	$frames = GGI_AUTO unless defined($frames);
	my($graphtype) = shift();
	$graphtype = GT_AUTO unless defined($graphtype);
	$graphtype = $graphtype->graphtype()
	    if ref($graphtype) 
		and UNIVERSAL::isa($graphtype, 'GGI::GraphType');
	$res = $this->_set_simple_mode($this->{obj}, 
				       $visx, $visy, $frames, $graphtype);
	delete($this->{pfmt}) unless $res; # Force new ggi_pixel_format alloc
	return $res;
    }
    # More than four parameters, get virtual size
    my($virtx, $virty);
    $arg = shift();
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($virtx, $virty) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($virtx, $virty) = ($arg, shift());
    }
    $virtx = GGI_AUTO unless defined($virtx);
    $virty = GGI_AUTO unless defined($virty);
    $arg = shift();
    if ($items == 5) {
	$arg = GT_AUTO unless defined($arg);
	$arg = $arg->graphtype() 
	    if ref($arg) and UNIVERSAL::isa($arg, 'GGI::GraphType');
	$res = $this->_set_graph_mode($this->{obj}, 
			       $visx, $visy, $virtx, $virty, $arg);
	delete($this->{pfmt}) unless $res; # Force new ggi_pixel_format alloc
	return $res;
    }
    # More than five parameters, get character cell size
    my($cellx, $celly);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($cellx, $celly) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($cellx, $celly) = ($arg, shift());
    }
    $cellx = GGI_AUTO unless defined($cellx);
    $celly = GGI_AUTO unless defined($celly);
    $arg = shift();
    $arg = GT_AUTO unless defined($arg);
    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::GraphType')) {
	$res = $this->_set_text_mode($this->{obj}, 
				     $visx, $visy, $virtx, $virty, 
				     $cellx, $celly, $arg->graphtype);
	delete($this->{pfmt}) unless $res; # Force new ggi_pixel_format alloc
	return $res;
    }
    $res = $this->_set_text_mode($this->{obj}, 
				 $visx, $visy, $virtx, $virty, 
				 $cellx, $celly, $arg);
    delete($this->{pfmt}) unless $res; # Force new ggi_pixel_format alloc
    return $res;
}

=head2 mode

The mode method returns a new GGI::Mode object representing the 
mode of the GGI:: object for which it is invoked, or a regular perl scalar
containing an error code if for some reason the object has no mode.
If a GGI::Mode object is returned, its "status" method is set to GGI_OK
(zero) and the GGI::Mode object's "lastused" method is set to point to the 
GGI:: object.  Note that a GGI::Mode object will return the value of
its "status" method when evaluated in a numerical or boolean 
context.

=cut

sub mode {
    my ($this) = shift();
    my($result) = new GGI::Mode;
    $this->_get_mode($this->{obj}, $result->{obj});
    return($result);
}

=head2 getmode

The getmode method fills the proferred GGI::Mode object with values
representing the mode of the GGI:: object's visual, and returns a
pointer to the same object, or a regular perl scalar containing an 
error code if for some reason the object has no mode.  If called 
with no parameters it is identical to the mode method and a new 
GGI::Mode object is constructed and returned.  If the proferred 
variable is not a GGI::Mode object, a new GGI::Mode object will be 
constructed and assigned to it.  Iff the getmode was sucessfull, 
the GGI::Mode object's "status" method is set to GGI_OK (zero) and the 
GGI::Mode object's "lastused" method is set to point to the GGI:: 
object.  Note that a GGI::Mode object will return the value of
its "status" method when evaluated in a numerical or boolean 
context.

=cut

sub getmode {
    my(@parms) = @_;
    my ($this) = shift(@parms);
    my ($res, $mode);
    if(not defined($_[1])) {
	$mode = new GGI::Mode;
        return $res if $res = $this->_get_mode($this->{obj}, $mode->{obj});
	$mode->status(0);
	$mode->lastused($this);
	return $mode;
    }
    unless (ref($_[1]) and UNIVERSAL::isa($_[1], 'GGI::Mode')) {
	$_[1] = new GGI::Mode;
        return $res if $res = $this->_get_mode($this->{obj}, $_[1]{obj});
	$_[1]->status(0);
	$_[1]->lastused($this);
	return $_[1];
    }
    return $res if $res = $this->_get_mode($this->{obj}, $_[1]{obj});
    $_[1]->status(0);
    $_[1]->lastused($this);
    return $_[1];
}

=head2 fg

This method sets or gets the GGI:: object's visual's graphics
context's pixel value for foreground drawing.  It takes either
a GGI::Color object or a single perl scalar containing a 
pixel value.  It returns the pixel value set (or if no
argument was provided, the value which was already set)
as a perl scalar.

=cut

sub fg {
    my($this) = shift();
    if (@_ + 0) {
	my($arg) = shift();
	return $this->_set_fg($this->{obj}, 
			      $this->_map($this->{obj}, $arg->{obj}, 0)) 
	    if ref($arg) and UNIVERSAL::isa($arg, 'GGI::Color');
	return $this->_set_fg($this->{obj}, $arg);
    }
    return $this->_get_fg($this->{obj}); 
}

=head2 fg

This method sets or gets the GGI:: object's visual's graphics
context's pixel value for background drawing.  It takes either
a GGI::Color object or a single perl scalar containing a 
pixel value.  It returns the pixel value set (or if no
argument was provided, the value which was already set)
as a perl scalar.

=cut

sub bg {
    my($this) = shift();
    if (@_ + 0) {
	my($arg) = shift();
	return $this->_set_bg($this->{obj}, 
			      $this->_map($this->{obj}, $arg->{obj}, 0)) 
	    if ref($arg) and UNIVERSAL::isa($arg, 'GGI::Color');
	return $this->_set_bg($this->{obj}, $arg);
    }
    return $this->_get_bg($this->{obj}); 
}

=head1 clip

This method sets the GGI:: object's visual's graphics
context's clipping rectangle.  The arguments may be perl scalars 
or GGI::Coord objects, specified in the order left-top, right-bottom.
The return value is a scalar containing a LibGGI error code,
zero on sucess, except when the method is called without arguments,
in which case it returns the clipping rectangle edges as an array of 
scalars.

=cut

sub clip {
    my($this) = shift();
    my($l, $t, $r, $b);
    unless (@_ + 0) {
	my($res) = $this->_get_clip($this->{obj}, $l, $t, $r, $b);
	return $res if $res;
	return ($l, $t, $r, $b);
    }
    ($l, $t, $r, $b) = @_;
    $l = defined($l) ? $l : 0;
    $t = defined($t) ? $t : 0;
    ($b, $r, $t, $l) = ($r, $t, $l->y(), $l->x())
	if (ref($l) and UNIVERSAL::isa($l, 'GGI::Coord'));
    ($b, $r) = ($r->y(), $r->x())
	if (ref($r) and UNIVERSAL::isa($r, 'GGI::Coord'));
    if (not defined($r) or not defined($b)) {
	my($mode) = $this->mode();
	return(int($mode)) if $mode; # error getting mode
	$r = defined($r) ? $r : $mode->virt->x;
	$b = defined($b) ? $b : $mode->virt->y;
    } 
    return $this->_set_clip($this->{obj}, $l, $t, $r, $b);
}

=head2 palette

When called with no arguments, results in a call to ggiSetColorfulPalette.

Other than this case, if the last parameter is not a reference, the palette 
is not altered, and the value of the current palette is fetched as per the 
GGI::Color Package's "get" method, with the exception that a GGI::Color object
is returned in a scalar context.

When called with an argument list containing either a GGI::Color object,
or a reference to an array, the function takes arguments as per the "set" 
method in the GGI::Color package (and will alter only those channels of 
the pallette which are selected by the channel mask, which defaults to 
'rgba').  See the GGI::Color documentation.

(IMPLEMENTATION NOTE: currently multiple 'slices' are not supported at all).

The contents of the proferred array or GGI::Color object are never altered.

In a scalar context, on success, a new GGI::Color object is returned, 
loaded with the current palette.  On failure a perl scalar with an error 
code is returned.  In an array context, undef() is returned if there is 
an error, otherwise an array is returned as per the GGI::Color get method.

=cut

sub palette {
    my($this) = shift();
    my($res, $col);
    unless (@_ + 0) {
	$res = $this->_color_pal($this->{obj});
	return undef unless defined(wantarray());
	return($res) if ($res);
    };
    unless (ref($_[-1]) and (UNIVERSAL::isa($_[-1], 'GGI::Color') or UNIVERSAL::isa($_[-1], 'ARRAY'))) {
	$col = new GGI::Color $this->_pal_len($this->{obj});
	$this->_get_pal($this->{obj}, 0, $col->size, $col->{obj});
	return $col->get(@_) if wantarray();
	return undef unless defined(wantarray());
	# This is the laziest implementation; multiple slices not implemented yet + should not go through perl array.
	my(@got) = $col->get(@_);
	my($csel, $idx, $len);
	$csel = shift();
	$csel = 'rgba' unless defined($csel);
	$idx = shift();
	$len = shift();
	$len = defined($len) ? ($len * length($csel)) : (@got + 0);
	$idx *= length($csel);
	@got = reverse(@got) if $idx < 0;
	$idx = abs($idx) - 1 if $idx < 0;
	while ($idx--) {
	    shift(@got);
	    $len--;
	}
	$res = new GGI::Color 
	    int(abs($len) / length($csel) + (abs($len) % length($csel) != 0)) ;
	$res->set($csel, 0, $len, \@got);
	return($res);
    };
    my(@newargs) = @_;
    # this is also a lazy and incorrect implementation.
    my ($arg);
    foreach $arg (@newargs) {
	$arg = [ $arg->get() ]
	    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Color'));
    }
    $col = new GGI::Color $this->_pal_len($this->{obj});
    $res = $this->_get_pal($this->{obj}, 0, $col->size, $col->{obj});
    return $res if $res;
    $col->set(@newargs);
    $res = $this->_set_pal($this->{obj}, 0, $col->size, $col->{obj});
    return $res if $res;
    return $col;
}

=head2 convert

This method converts data between GGI::Color objects and GGI::Pixel
objects (or a single pixel values carried in perl scalars).  

If the last argument is a GGI::Color object, the function returns
a new GGI::Pixel object containing the pixel values corresponding
to the colors contained in that object.  The GGI::Color object
may be preceded by a index at which to start within the GGI::Color
object, and, optionally, a number of color values to convert.

Otherwise, normal perl scalars and GGI::Pixel objects may be mixed
and a new GGI::Color object is returned containing the colors
that correspond to the pixel values.

=cut

sub convert {
    my ($this) = shift();
    my($arg) = @_[-1];
    my($res, $len, $idx);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Color')) {
	my($col, $start, $len);
	$col = pop();
	$len = pop();
	$start = pop();
	$len = $col->size unless defined($len);
	$start = 0 unless $start > 0;
	$res = new GGI::Pixel $this $len;
	$this->_pack($this->{obj}, $res->{buf}, $col->{obj}, $start, $len);
	return $res;
    }
    else {
	foreach $arg (@_) {
	    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
		$len += $arg->size;
	    }
	    else {
		$len++;
	    }
	    $res = new GGI::Color $len;
	}
	return undef unless $len;
	foreach $arg (@_) {
	    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
		$this->_unpack($this->{obj}, $arg->{buf}, 
			       $res->{obj}, $idx, $arg->size);
		$idx += $arg->size;
	    }
	    else {
		$res->_unmap($this->{obj}, $arg, $res->{obj}, $idx);
		$idx++;
	    }
	}
    }
    return $res;
}

# These are not implemented yet:
#  $num_mapped = $visual->map($pixels, $colors);
#  $num_mapped = $visual->map($pixel, $colors);
#  $num_mapped = $visual->map($pixels, $start_color, $colors);
#  $num_mapped = $visual->map($pixels, $start_color, $num, $colors);
#  $num_unmapped = $visual->unmap($colors, $pixels);
#  $num_unmapped = $visual->unmap($colors, $pixel);
#  $num_unmapped = $visual->unmap($start_color, $colors, $pixels);
#  $num_unmapped = $visual->unmap($start_color, $num, $colors, $pixels);

=head2 pixel

This method is used to set an individual pixel in the GGI:: object's visual
at a given coordinate.  With no additional arguments, the pixel is set to the
GGI:: object's visual's graphics context's pixel value for foreground drawing.

If the first argument to the pixel method is a GGI::Pixel object, then
the value the pixel is instead retreived and placed in the GGI::Pixel object.

Alternatively, if the last argument to the pixel method is a 
GGI::Pixel object, then the pixel is set to the value of that GGI::Pixel
object.

The pixel method returns a LibGGI error code as a perl scalar, GGI_OK
(zero) means the operation was successful.

=cut

sub pixel {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
	my($dest) = $arg;
	$dest->lastused($this);
	$dest->resize_get(1);
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    ($x, $y) = ($arg->x, $arg->y);
	}
	else {
	    ($x, $y) = ($arg, shift());
	}
	return $this->_get_pixel($this->{obj}, $x, $y, $dest->{buf});
    }
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    $arg = (@_ + 0) ? shift() : undef;
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI')) {
        # Crossblit -- call other visual's method.
	my($that) = $arg;
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    return $that->_crossblit($that->{obj}, $arg->x, $arg->y, 1, 1, 
				     $this->{obj}, $x, $y);
	}
	return $that->_crossblit($that->{obj}, $arg, shift(), 1, 1, 
			    $this->{obj}, $x, $y);
    }	
    elsif (defined($arg)) {
	return $this->_put_pixel($this->{obj}, $x, $y, $arg)
	    unless ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel') ;
	$arg->lastused($this);
	$arg->resize_put(1);
	return $this->_put_pixel_buf($this->{obj}, $x, $y, $arg->{buf});
    }
    else {
	return $this->_draw_pixel($this->{obj}, $x, $y);
    }
}

=head2 hline

The hline method allows for manipulating a row of contiguous pixels
in the GGI:: object's visual.

If the first argument is a GGI::Pixel object, pixel data is taken
from the visual into the GGI::Pixel object's buffer.  In this
case the next parameter is a perl scalar designating how many pixels 
to take, and this is followed by the coordinates of the leftmost
point on the desired line, either as two perl scalars or as  a
single GGI::Coord object.

Otherwise, the coordinates of the leftmost point on the desired line
precede the width parameter.  In this case, if a GGI::Pixel object 
appears as the last argument, pixel data is taken from the GGI::Pixel
object's buffer and written to the visual.  If the GGI::Pixel parameter
is not present, the line is filled with the GGI:: object's visual's
graphics context's foreground drawing color.

=cut

sub hline {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
	my($dest) = $arg;
	$dest->lastused($this);
	my ($w) = shift();
	$dest->resize_get($w);
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    ($x, $y) = ($arg->x, $arg->y);
	}
	else {
	    ($x, $y) = ($arg, shift());
	}
	return $this->_get_hline($this->{obj}, $x, $y, $w, $dest->{buf});
    }
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    my($w) = shift();
    $arg = (@_ + 0) ? shift() : undef;
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI')) {
        # Crossblit -- call other visual's method.
	my($that) = $arg;
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    return $that->_cross_blit($that->{obj}, $arg->x, $arg->y, $w, 1, 
				      $this->{obj}, $x, $y);
	}
	return $that->_cross_blit($that->{obj}, $arg, shift(), $w, 1, 
				  $this->{obj}, $x, $y);
    }	
    elsif (defined($arg)) {
	$arg->lastused($this);
	$arg->resize_put($w);
	return $this->_put_hline($this->{obj}, $x, $y, $w, (shift())->{buf});
    }
    else {
	return $this->_draw_hline($this->{obj}, $x, $y, $w);
    }
}

=head2 vline

The vline method allows for manipulating a column of contiguous pixels
in the GGI:: object's visual.

If the first argument is a GGI::Pixel object, pixel data is taken
from the visual into the GGI::Pixel object's buffer.  In this
case the next parameter is a perl scalar designating how many pixels 
to take, and this is followed by the coordinates of the topmost
point on the desired line, either as two perl scalars or as  a
single GGI::Coord object.

Otherwise, the coordinates of the topmost point on the desired line
precede the height parameter.  In this case, if a GGI::Pixel object 
appears as the last argument, pixel data is taken from the GGI::Pixel
object's buffer and written to the visual.  If the GGI::Pixel parameter
is not present, the line is filled with the GGI:: object's visual's
graphics context's foreground drawing color.

=cut

sub vline {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
	my($dest) = $arg;
	$dest->lastused($this);
	my ($h) = shift();
	$dest->resize($h, "get");
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    ($x, $y) = ($arg->x, $arg->y);
	}
	else {
	    ($x, $y) = ($arg, shift());
	}
	return $this->_get_vline($this->{obj}, $x, $y, $h, $dest->{buf});
    }
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    my($h) = shift();
    $arg = (@_ + 0) ? shift() : undef;
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI')) {
        # Crossblit -- call other visual's method.
	my($that) = $arg;
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    return $that->_cross_blit($that->{obj}, $arg->x, $arg->y, 1, $h, 
				      $this->{obj}, $x, $y);
	}
	return $that->_cross_blit($that->{obj}, $arg, shift(), 1, $h,
				  $this->{obj}, $x, $y);
    }	
    elsif (defined($arg)) {
	$arg->lastused($this);
	$arg->resize_put($h);
	return $this->_put_vline($this->{obj}, $x, $y, $h, (shift())->{buf});
    }
    else {
	return $this->_draw_vline($this->{obj}, $x, $y, $h);
    }
}

=head2

The line method draws a line between two points.  The points
may each be given as GGI::Coord objects or as two perl scalars.
The line is drawn in the GGI:: object's visual's graphic context's
foreground drawing color.

=cut

sub line {
    my($this) = shift();
    my($arg) = shift();
    my($x1, $y1, $x2, $y2);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x1, $y1) = ($arg->x, $arg->y);
    }
    else {
	($x1, $y1) = ($arg, shift());
    }
    $arg = shift();
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x2, $y2) = ($arg->x, $arg->y);
    }
    else {
	($x2, $y2) = ($arg, shift());
    }
    return($this->_draw_line($this->{obj},$x1,$y1,$x2,$y2));
}

=head2 box

The box method allows for manipulating a rectangle of contiguous pixels
in the GGI:: object's visual.

If the first argument is a GGI::Pixel object, pixel data is taken
from the visual into the GGI::Pixel object's buffer.  In this
case the next parameters are a width and height, and the coordinates
of the top-left corner of the box.  Both pairs of parameters may
be represented by either a pair of perl scalars or a GGI::Coord 
object.

Otherwise, the coordinates of the top-left pixel of the box
precede the height/width parameter(s).  In this case, if a GGI::Pixel 
object appears as the last argument, pixel data is taken from the GGI::Pixel
object's buffer and written to the visual.  If the GGI::Pixel parameter
is not present, the line is filled with the GGI:: object's visual's
graphics context's foreground drawing color.  If a GGI:: object appears
after the coordinates, the box is copied from that GGI:: object's
visual.  Coordinates must follow the GGI:: object designating the
top-left corner of the box from which to take the pixel data.

=cut

sub box {
    my($this) = shift();
    my($arg) = shift();
    my($w, $h);
    if (ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
	my($dest) = $arg;
	$dest->lastused($this);
	my ($w, $h);
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    ($w, $h) = ($arg->x, $arg->y);
	}
	else {
	    ($w, $h) = ($arg, shift());
	}
	$dest->resize_get($w * $h);
	$arg = shift();
	my($srcx, $srcy);
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    ($srcx, $srcy) = ($arg->x, $arg->y);
	}
	else {
	    ($srcx, $srcy) = ($arg, shift());
	}
	return $this->_get_box($this->{obj}, $srcx, $srcy, 
			       $w, $h, $dest->{buf});
    }
    my($destx, $desty);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($destx, $desty) = ($arg->x, $arg->y);
    }
    else {
	($destx, $desty) = ($arg, shift());
    }
    $arg = shift();
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($w, $h) = ($arg->x, $arg->y);
    }
    else {
	($w, $h) = ($arg, shift());
    }
    return ($this->_draw_box($this->{obj}, $destx, $desty, $w, $h)) unless (@_ + 0);
    $arg = shift();
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Pixel')) {
	$arg->lastused($this);
	$arg->resize_put($w * $h);
	return $this->_put_box($this->{obj}, $destx, $desty, 
			       $w, $h, $arg->{buf});
    }
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI')) { 
        # Crossblit -- call other visual's method.
	my($that) = $arg;
	$arg = shift();
	if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	    return $that->_cross_blit($that->{obj}, $arg->x, $arg->y, $w, $h, 
				      $this->{obj}, $destx, $desty);
	}
	return $that->_cross_blit($that->{obj}, $arg, shift(), $w, $h, 
				  $this->{obj}, $destx, $desty);
    }
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	return $this->_copy_box($this->{obj},$arg->x, $arg->y, $w, $h, 
				$destx, $desty);
    }
    return $this->_copy_box($this->{obj}, $arg, shift(), $w, $h, 
			    $destx, $desty);
}

=head2

The flush method, when called with no parameters, will cause any
drawing operations which have not yet been displayed on the visual
to be flushed.  It may also be called with the coordinates of a box
designating a region of the visual which should be flushed.  The box
parameters (top-left corner and width/height in that order) may each
be given as GGI::Coord objects or as two perl scalars.

=cut

sub flush {
    my($this) = shift();
    $this->_flush($this->{obj}) unless (@_ + 0);
    my($arg) = shift();
    my($x, $y, $w, $h);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    $arg = shift();
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($w, $h) = ($arg->x, $arg->y);
    }
    else {
	($w, $h) = ($arg, shift());
    }
    return $this->_flush_region($this->{obj}, $x, $y, $w, $h);
}

=head2

The puts method is intended for debugging or quick-hack use only, 
not as full featured text engine.  It draws text at coordinates given 
either a GGI::Coord object or two perl scalars.  The text to be drawn
is passed as a perl scalar as the last parameter, and is evaluated
in a string context before it is rendered.

=cut

sub puts {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if(ref($arg) and UNIVERSAL::isa($arg, 'GGI::Coord')) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    $arg = shift();
    return $this->_put_s($this->{obj}, $x, $y, $arg);
}

=head2

The stringify method creates a string containing debugging information
about the GGI:: object.  It overloads the text stringification operator
for the GGI:: class, so there is little need to call it directly.

=cut

sub stringify {
    my ($this) = shift();
    return 
	sprintf "Visual at %p (display params: %s)", 
	$this->{obj}, join(" ", @{$this->{display}});
}

=head2 send

The send method will inject the proferred GII::Event object into 
the GGI:: object's gii_input.

=cut

# send method inherited, C function overloaded to use ggiEventSend

=head2 read

Without an argument, waits for (blocks) and returns the least recent 
event received on the gii_input represented by the invoking object.
An optional first argument can be used to set an event mask,
such that events that do not match the mask are left untouched 
and not returned.  The returned object is a new GII::Event object,
and has been subclassed into the correct GII::Event subclass.
The function may also return undef if there was an error.

A second argument allows you to reuse a GII::Event, the new event
is placed in this object.  In this case, using a first argument of
undef() will force a pass-all event mask.  Note you should take
care when re-using events; it is assumed you know what you are
doing if you try this.  The return value is the same as above in this
case.

Event mask constants may be imported by explicitly calling use()
on GII::Event, or they may be referred to by fully qualified
identifiers (they are in the GII:: namespace as well as the GII::Event::
namespace.)

=cut

# read method inherited, C function overloaded to use ggiEventRead

=head2 poll

Without an argument, returns a mask telling which types of events 
are waiting to be read from the input associated with the GGI:: 
object.

An optional first argument can be a perl scalar representing
a maximum number of microseconds to wait.  (IMPLEMENTATION
NOTE: will eventually add timeval and Time:: objects to acceptable 
parameters.)

An optional second argument can be used to set an event mask,
such that events that do not match the mask are not checked for
or represented in the return value.

Event mask constants may be imported by explicitly calling use()
on GII::Event, or they may be referred to by fully qualified
identifiers (they are in the GII:: namespace as well as the GII::Event::
namespace.)

=cut

# poll method inherited, C function overloaded to use ggiEventPoll

=head2 queued

Returns the number of events waiting to be read on the input
associated with the GGI:: object.

=cut

# queued method inherited, C function overloaded to use ggiEventsQueued

=head2 mask

Without an argument, gets the mask which is filtering events on
the input associated with the GGI:: object (events not matching the 
mask are discarded from the input or simply not generated in the first place.)

With an argument, sets the mask to the given value, returning
an error code or zero (GGI_OK) for success.

=cut

# mask method inherited, C function overloaded to use ggi(Get|Set)EventMask

# _get_pixel_format is an internal method to get a raw reference to 
# a copy of the ggi_pixelformat structure.  It will reuse the last copy 
# if the mode has not changed, or retreive, alloc, and copy a new one if
# not, stuffing it in a blessed reference for housekeeping.

sub _get_pixel_format {
    my($this) = shift();
    return $this->{pfmt} if exists($this->{pfmt});
    my ($newpfmt) = $this->_clone_pixel_format($this->{obj});
    $this->{pfmt} = \$newpfmt;
    return bless $this->{pfmt}, 'GGI::Pixel::_internal';
}

END {
    die "Error De-Initializing LibGGI!  Sucks to be you." if ggiExit();
}

1;
__END__


=head1 AUTHOR

Brian S. Julin <bri@calyx.com>

=head1 SEE ALSO

GII::(3pm), perl(1).

=cut
